<?php

namespace FiloBlu\Flow\Model\From\Entity;

use Magento\Framework\ObjectManagerInterface;

/**
 * Class AbstractEntity
 * @package FiloBlu\Flow\Model\From\Entity
 */
abstract class AbstractEntity
{
    /**
     * @var string
     */
    protected $specialAttributeRex = '^_';
    /**
     * @var mixed|null
     */
    protected $_entityTypeCode = null;
    /**
     * @var null
     */
    protected $_entityTypeModel = null;
    /**
     * @var mixed|null
     */
    protected $_entityIdentifier = null;
    /**
     * @var string|null
     */
    protected $_attributeCode = null;
    /**
     * @var string|null
     */
    protected $_attributeValue = null;
    /**
     * @var mixed|null
     */
    protected $_storeId = null;
    /**
     * @var null
     */
    protected $attributeModel = null;
    /**
     * @var null
     */
    protected $_objectEntity = null;
    /**
     * @var null
     */
    protected $_objectEntityId = null;
    /**
     * @var \Magento\Catalog\Model\ProductFactory
     */
    protected $_product_factory;
    /**
     * @var \Magento\Framework\Registry
     */
    protected $_registry;
    /**
     * @var \FiloBlu\Flow\Helper\Attribute
     */
    protected $helper_attribute;
    /**
     * @var \FiloBlu\Flow\Helper\Monolog\Logger|\Monolog\Logger
     */
    protected $logger;
    /**
     * @var ObjectManagerInterface
     */
    protected $objectManager;

    /**
     * @var \FiloBlu\Flow\Helper\Product
     */
    protected $helperProduct;

    /**
     * @var \Magento\Eav\Model\Entity\AttributeFactory
     */
    protected $attributeFactory;

    /**
     * @var \Magento\Eav\Model\ResourceModel\Entity\Attribute\Option\CollectionFactory
     */
    protected $attributeOptionCollectionFactory;

    /**
     * AbstractEntity constructor.
     * @param \Magento\Framework\Registry $registry
     * @param \Magento\Catalog\Model\ProductFactory $productFactory
     * @param \FiloBlu\Flow\Helper\Product $helper_product
     * @param \FiloBlu\Flow\Helper\Attribute $helper_attribute
     * @param \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider
     * @param \Magento\Eav\Model\Entity\AttributeFactory $attributeFactory
     * @param \Magento\Eav\Model\ResourceModel\Entity\Attribute\Option\CollectionFactory $attributeOptionCollectionFactory
     * @param ObjectManagerInterface $objectManager
     * @param array $data
     * @throws \Exception
     */
    public function __construct(

        \Magento\Framework\Registry $registry,
        \Magento\Catalog\Model\ProductFactory $productFactory,
        \FiloBlu\Flow\Helper\Product $helper_product,
        \FiloBlu\Flow\Helper\Attribute $helper_attribute,
        \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider,
        \Magento\Eav\Model\Entity\AttributeFactory $attributeFactory,
        \Magento\Eav\Model\ResourceModel\Entity\Attribute\Option\CollectionFactory $attributeOptionCollectionFactory,
        ObjectManagerInterface $objectManager,
        array $data = []
    )
    {
        $this->attributeOptionCollectionFactory = $attributeOptionCollectionFactory;
        $this->attributeFactory = $attributeFactory;
        $this->helperProduct = $helper_product;
        $this->objectManager = $objectManager;
        $this->_storeId = $data['storeId'];
        $this->_attributeCode = trim($data['attributeCode']); // a
        $this->_attributeValue = trim($data['attributeValue']); // v
        $this->_entityIdentifier = $data['entityIdentifier']; // e
        $this->_entityTypeCode = $data['entityTypeCode']; // valori: catalog_product, product_attribute
        $this->_product_factory = $productFactory;
        $this->_registry = $registry;
        $this->helper_attribute = $helper_attribute;
        $this->getEntityModel();
        $this->logger = $loggerProvider->getLogger();
    }

    /**
     * @return |null
     * @throws \Exception
     */
    protected function getEntityModel()
    {
        if (is_null($this->_entityTypeModel)) {
            $entityType = $this->objectManager->create('Magento\Eav\Model\Entity\Type')->loadByCode($this->getEntityTypeCode());
            if (!$entityType->getId()) {
                throw new \Exception('EntityType not found for code ' . $this->_entityTypeCode);
            }

            $this->_entityTypeModel = $entityType->getEntityModel();
        }

        return $this->_entityTypeModel;
    }

    /**
     * @return mixed|null
     * @throws \Exception
     */
    public function getEntityTypeCode()
    {
        if (empty($this->_entityTypeCode)) {
            throw new \Exception('EntityTypeCode was not defined for this class');
        }

        return $this->_entityTypeCode;
    }

    /**
     * @throws \Exception
     */
    public function process()
    {
        if ($this->isSpecialFunction($this->_attributeCode)) {
           $this->processSpecialFunction();
        } else {
            $this->setObjectAttributeValue();
        }
    }

    /**
     * @param $attribute
     * @return false|int
     */
    protected function isSpecialFunction($attribute)
    {
        return preg_match('/' . $this->specialAttributeRex . '/', $attribute);
    }

    /**
     * @return mixed
     * @throws \Exception
     */
    protected function processSpecialFunction()
    {
        $className = $this->getSpecialActionClassName();

        if (class_exists($className)) {
            $specificSpecialClass = $this->objectManager->create($className);

            if (!method_exists($specificSpecialClass, 'process')) {
                throw new \Exception('Missing special function for attribute ' . $this->_attributeCode);
            }
            return $specificSpecialClass->process($this->getObject(), $this->_attributeValue, $this->_storeId);
        } else {
            throw new \Exception("A special class \"{$className}\" for attribute code (\"{$this->_attributeCode}\") was not found");
        }

    }

    /**
     * @return string
     */
    protected function getSpecialActionClassName()
    {
        return '\FiloBlu\Flow\Model\Actions\From\\' . ucfirst(str_replace("_", "", $this->_attributeCode));
    }

    /**
     * @return |null
     * @throws \Exception
     */
    protected function getObject()
    {
        if (is_null($this->_objectEntity)) {
            if (!$this->getObjectId()) {
                throw new \Exception('No id found for model ' . $this->getEntityModel() . ' with given identifier ' . $this->_entityIdentifier);
            }
            $_prod_sku = $this->_product_factory->create()->getIdBySku($this->_entityIdentifier);
            $this->_objectEntity = $this->_product_factory->create()->setStoreId($this->_storeId)->load($_prod_sku);
            if (!$this->_objectEntity->getId()) {
                throw new \Exception('No instance found for model ' . $this->getEntityModel() . ' with id ' . $this->getObjectId());
            }
        }
        return $this->_objectEntity;
    }

    /**
     * @return mixed|null
     */
    public function getObjectId()
    {
        if (is_null($this->_objectEntityId)) {
            $this->_objectEntityId = $this->_entityIdentifier;
        }
        return $this->_objectEntityId;
    }

    /**
     * @throws \Exception
     */
    protected function setObjectAttributeValue()
    {
        $object = $this->getObject();
        $attributeValue = $this->getAttributeValue();
        $input_type = strtolower($this->getAttribute()->getFrontend()->getInputType());
        try {
            if ($input_type === 'multiselect' && empty($attributeValue)) {
                $object->unsetData($this->_attributeCode);
                $object->save();
            } elseif (is_null($attributeValue) || empty($attributeValue)) {
                $object->unsetData($this->_attributeCode);
            } else {
                $object->setDataUsingMethod($this->_attributeCode, $attributeValue);
                $object->getResource()->saveAttribute($object, $this->_attributeCode);
            }
        } catch (\Exception $e) {
            $message = "Problem while processing eavs, <br/> store id : " . $this->_storeId . "<br/> id : "
                . $object->getId() . "<br/> attribute data : " . $attributeValue;
            $message .= "<br/> \Exception : " . $e->getMessage();
            $message .= "<br/> Backtrace : <br/> " . nl2br($e->getTraceAsString());
            throw new \Exception($e->getMessage());
        }
    }

    /**
     * @return mixed|string|null
     * @throws \Exception
     */
    protected function getAttributeValue()
    {

        $attribute_model = $this->getAttribute();
        $attributeValue = $this->_attributeValue;
        $input_type = strtolower($attribute_model->getFrontend()->getInputType());

        if (strlen($attributeValue) > 0 && ($input_type === 'select' || $input_type === 'multiselect') && $this->getAttributeCode() != 'visibility' && $this->getAttributeCode() != 'country_of_manufacture') {
            $attributeValueLabels = [];
            $attributeValueArray = [];
            if ($input_type === 'multiselect') {
                if (strpos($attributeValue, '|') !== false) {
                    //if there is | as separator, we need to replace with ,
                    $attributeValue = str_replace('|', ',', $attributeValue);
                }
                $attributeValueLabels = explode(',', $attributeValue);
                $attributeValueLabels = array_map('trim', $attributeValueLabels);
                $attributeValueLabels = array_unique($attributeValueLabels);
            } else {
                $attributeValueLabels[] = $attributeValue;
            }
            foreach ($attributeValueLabels as $attributeValueLabel) {
                $optionId = $this->getAttributeOptionValue($attributeValueLabel);
                if (empty($optionId) && $this->_storeId == 0) { // && Mage::getStoreConfig('fbflow/new_options_status') == 1) {
                    $this->addAttributeOptionValue($attributeValueLabel);
                    $optionId = $this->getAttributeOptionValue($attributeValueLabel);
                }
                if (empty($optionId)) {
                    throw new \Exception('Couldn\'t find nor add option ' . $attributeValueLabel . ' for attribute ' . $this->_attributeCode . ' entity type ' . $this->getEntityTypeCode() . ' on store ' . $this->_storeId);
                }
                $attributeValueArray[] = $optionId;
            }
            $attributeValue = implode(',', $attributeValueArray);
        }

        return $attributeValue;
    }

    /**
     * @return |null
     * @throws \Exception
     */
    protected function getAttribute()
    {
        if (empty($this->attributeModel)) {
            $this->attributeModel = $this->attributeFactory->create()->loadByCode($this->getEntityTypeCode(), $this->getAttributeCode());
            if (!$this->attributeModel->getId()) {
                throw new \Exception("No attribute found for code {$this->_attributeCode}");
            }
        }
        return $this->attributeModel;
    }

    /**
     * @return string|null
     * @throws \Exception
     */
    public function getAttributeCode()
    {
        if (empty($this->_attributeCode)) {
            throw new \Exception('AttributeCode was not defined for this class');
        }

        return $this->_attributeCode;
    }

    /**
     * @param $attributeValueLabel
     * @return mixed
     */
    protected function getAttributeOptionValue($attributeValueLabel)
    {

        $collection = $this->attributeOptionCollectionFactory->create();

        $collection->getSelect()
            ->join('eav_attribute', 'eav_attribute.attribute_id=main_table.attribute_id', 'eav_attribute.attribute_code')
            ->join('eav_attribute_option_value', 'eav_attribute_option_value.option_id=main_table.option_id', 'value');

        $collection
            ->addFieldToFilter('eav_attribute_option_value.value', ['eq' => $attributeValueLabel])
            ->addFieldToFilter('attribute_code', ['eq' => $this->_attributeCode])
            ->addFieldToFilter('store_id', ['eq' => $this->_storeId]);

        $attribute = $collection->getFirstItem();

        return $attribute->getOptionId();
    }

    /**
     * @param $attributeValueLabel
     * @return bool
     * @throws \Exception
     */
    protected function addAttributeOptionValue($attributeValueLabel)
    {
        try {

            return $this->helper_attribute->addAttributeOption(
                [
                    'attribute_id' => $this->getAttribute()->getId(),
                    'store_id' => 0,
                    'value' => $attributeValueLabel
                ]
            );

        } catch (\Exception $e) {
            throw new \Exception('Couldn\'t add option ' . $attributeValueLabel . ' for attribute ' . $this->_attributeCode . ' entity type ' . $this->getEntityTypeCode());
        }

    }

}
